#!/usr/bin/env python
# -*- coding: utf-8 -*-

# standalone - standalone ACME challenge webserver
# Copyright (c) Markus Hauschild & David Klaftenegger, 2016.
# Copyright (c) Rudolf Mayerhofer, 2019.
# available under the ISC license, see LICENSE

try:
    from BaseHTTPServer import HTTPServer, BaseHTTPRequestHandler
except ImportError:
    from http.server import HTTPServer, BaseHTTPRequestHandler

import re
import socket
import threading

from acertmgr.modes.webdir import HTTPChallengeHandler
from acertmgr.tools import log

HTTPServer.allow_reuse_address = True


class HTTPServer6(HTTPServer):
    address_family = socket.AF_INET6


class ChallengeHandler(HTTPChallengeHandler):
    def __init__(self, config):
        HTTPChallengeHandler.__init__(self, config)
        self.bind_address = config.get("bind_address", "")
        self.port = int(config.get("port", 80))

        self.challenges = {}  # Initialize the challenge data dict
        self.server_thread = None
        self.server = None

    def create_challenge(self, domain, thumbprint, token):
        self.challenges[token] = "{0}.{1}".format(token, thumbprint)

    def destroy_challenge(self, domain, thumbprint, token):
        del self.challenges[token]

    def start_challenge(self, domain, thumbprint, token):
        _self = self

        # Custom HTTP request handler
        class _HTTPRequestHandler(BaseHTTPRequestHandler):
            def log_message(self, fmt, *args):
                log("Request from '%s': %s" % (self.address_string(), fmt % args))

            def do_GET(self):
                # Match token on http://<domain>/.well-known/acme-challenge/<token>
                match = re.match(r'.*/(?P<token>[^/]*)$', self.path)
                if match and match.group('token') in _self.challenges:
                    value = _self.challenges[match.group('token')].encode('utf-8')
                    rcode = 200
                else:
                    value = "404 - NOT FOUND".encode('utf-8')
                    rcode = 404
                self.send_response(rcode)
                self.send_header('Content-type', 'text/plain')
                self.send_header('Content-length', len(value))
                self.end_headers()
                self.wfile.write(value)

        try:
            self.server = HTTPServer6((self.bind_address, self.port), _HTTPRequestHandler)
        except socket.gaierror:
            self.server = HTTPServer((self.bind_address, self.port), _HTTPRequestHandler)

        def _serve():
            self.server.serve_forever()

        self.server_thread = threading.Thread(target=_serve)
        self.server_thread.start()
        HTTPChallengeHandler.start_challenge(self, domain, thumbprint, token)

    def stop_challenge(self, domain, thumbprint, token):
        if self.server_thread.is_alive():
            self.server.shutdown()
            self.server_thread.join()
            self.server.server_close()
            self.server = None
            self.server_thread = None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# abstract - abstract base classes for challenge handlers
# Copyright (c) Rudolf Mayerhofer, 2019.
# available under the ISC license, see LICENSE


class AbstractChallengeHandler:
    def __init__(self, config):
        self.config = config

    @staticmethod
    def get_challenge_type():
        raise NotImplementedError

    def create_challenge(self, domain, thumbprint, token):
        raise NotImplementedError

    def destroy_challenge(self, domain, thumbprint, token):
        raise NotImplementedError

    # Optional: Indicate when a challenge request is imminent
    def start_challenge(self, domain, thumbprint, token):
        pass

    # Optional: Indicate when a challenge response has been received
    def stop_challenge(self, domain, thumbprint, token):
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# acertmgr - generic acme api functions
# Copyright (c) Markus Hauschild & David Klaftenegger, 2016.
# Copyright (c) Rudolf Mayerhofer, 2019.
# available under the ISC license, see LICENSE


class ACMEAuthority:
    # @brief Init class with config
    # @param config Configuration data
    # @param key Account key data
    def __init__(self, config, key):
        self.key = key
        self.config = config

    # @brief register an account over ACME
    def register_account(self):
        raise NotImplementedError

    # @brief function to fetch certificate using ACME
    # @param csr the certificate signing request in pyopenssl format
    # @param domains list of domains in the certificate, first is CN
    # @param challenge_handlers a dict containing challenge for all given domains
    # @return the certificate
    def get_crt_from_csr(self, csr, domains, challenge_handlers):
        raise NotImplementedError

    # @brief function to revoke a certificate using ACME
    # @param crt certificate to revoke
    # @param reason (int) optional certificate revoke reason (see https://tools.ietf.org/html/rfc5280#section-5.3.1)
    def revoke_crt(self, crt, reason=None):
        raise NotImplementedError

    # @brief check if the given certificate should be renewed according to it's RFC9773 data
    # @param crt certificate to check
    # @param issuer certificate necessary for correct AKI data
    # @return True if the certificate should be renewed, False otherwise
    def check_ari_for_renewal(self, crt, issuer):
        return False  # Do not mark for renewal if this is not implemented properly (use default renewal checks instead)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# modes - challenge handler modes package
# Copyright (c) Rudolf Mayerhofer, 2019.
# available under the ISC license, see LICENSE

import importlib
import json

challenge_handlers = dict()


# @brief find or create a challenge handler for the given settings
# @param settings the domain's configuration options
def challenge_handler(settings):
    key = json.dumps(settings, sort_keys=True)
    if key in challenge_handlers:
        return challenge_handlers[key]
    else:
        if "mode" in settings:
            mode = settings["mode"]
        else:
            mode = "standalone"

        handler_module = importlib.import_module("acertmgr.modes.{0}".format(mode))
        handler_class = getattr(handler_module, "ChallengeHandler")
        handler_obj = handler_class(settings)
        challenge_handlers[key] = handler_obj
        return handler_obj



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# acertmgr - acme api v1 functions
# Copyright (c) Markus Hauschild & David Klaftenegger, 2016.
# Copyright (c) Rudolf Mayerhofer, 2019.
# available under the ISC license, see LICENSE

import copy
import json
import re
import time

from acertmgr import tools
from acertmgr.authority.acme import ACMEAuthority as AbstractACMEAuthority
from acertmgr.tools import log


class ACMEAuthority(AbstractACMEAuthority):
    # @brief Init class with config
    # @param config Configuration data
    # @param key Account key data
    def __init__(self, config, key):
        log('You currently use ACMEv1 which is deprecated, consider using ACMEv2 (RFC8555) if at all possible.',
            warning=True)
        AbstractACMEAuthority.__init__(self, config, key)
        self.registered_account = False
        self.ca = config['authority']
        self.agreement = config['authority_tos_agreement']

    # @brief create the header information for ACME communication
    # @param key the account key
    # @return the header for ACME
    def _prepare_header(self):
        alg, jwk = tools.get_key_alg_and_jwk(self.key)
        header = {
            "alg": alg,
            "jwk": jwk,
        }
        return header

    # @brief helper function to make signed requests
    # @param url the request URL
    # @param header the message header
    # @param payload the message
    # @return tuple of return code and request answer
    def _send_signed(self, url, header, payload):
        payload64 = tools.bytes_to_base64url(json.dumps(payload).encode('utf8'))
        protected = copy.deepcopy(header)
        protected["nonce"] = tools.get_url(self.ca + "/directory").headers['Replay-Nonce']
        protected64 = tools.bytes_to_base64url(json.dumps(protected).encode('utf8'))
        out = tools.signature_of_str(self.key, '.'.join([protected64, payload64]))
        data = json.dumps({
            "header": header, "protected": protected64,
            "payload": payload64, "signature": tools.bytes_to_base64url(out),
        })
        try:
            resp = tools.get_url(url, data.encode('utf8'))
            return resp.getcode(), resp.read()
        except IOError as e:
            return getattr(e, "code", None), getattr(e, "read", e.__str__)()

    # @brief register an account over ACME
    # @return True if new account was registered, False otherwise
    def register_account(self):
        if self.registered_account:
            # We already have registered with this authority, just return
            return

        header = self._prepare_header()
        code, result = self._send_signed(self.ca + "/acme/new-reg", header, {
            "resource": "new-reg",
            "agreement": self.agreement,
        })
        if code == 201:
            log("Registered!")
            self.registered_account = True
            return True
        elif code == 409:
            log("Already registered!")
            self.registered_account = True
            return False
        else:
            raise ValueError("Error registering: {0} {1}".format(code, result))

    # @brief function to fetch certificate using ACME
    # @param csr the certificate signing request in pyopenssl format
    # @param domains list of domains in the certificate, first is CN
    # @param challenge_handlers a dict containing challenge for all given domains
    # @return the certificate and corresponding ca as a tuple
    # @note algorithm and parts of the code are from acme-tiny
    def get_crt_from_csr(self, csr, domains, challenge_handlers):
        header = self._prepare_header()
        account_thumbprint = tools.bytes_to_base64url(
            tools.hash_of_str(json.dumps(header['jwk'], sort_keys=True, separators=(',', ':'))))

        challenges = dict()
        tokens = dict()
        authdomains = list()
        # verify each domain
        try:
            for domain in domains:
                log("Verifying {0}...".format(domain))

                # get new challenge
                code, result = self._send_signed(self.ca + "/acme/new-authz", header, {
                    "resource": "new-authz",
                    "identifier": {"type": "dns", "value": domain},
                })
                if code != 201:
                    raise ValueError("Error requesting challenges: {0} {1}".format(code, result))

                # create the challenge
                authz = json.loads(result.decode('utf8'))
                if authz.get('status', 'no-status-found') == 'valid':
                    log("{} has already been verified".format(domain))
                    continue
                challenges[domain] = [c for c in authz['challenges'] if
                                      c['type'] == challenge_handlers[domain].get_challenge_type()][0]
                tokens[domain] = re.sub(r"[^A-Za-z0-9_\-]", "_", challenges[domain]['token'])

                if domain not in challenge_handlers:
                    raise ValueError("No challenge handler given for domain: {0}".format(domain))

                challenge_handlers[domain].create_challenge(domain, account_thumbprint, tokens[domain])
                authdomains.append(domain)

            # after all challenges are created, start processing authorizations
            for domain in authdomains:
                try:
                    challenge_handlers[domain].start_challenge(domain, account_thumbprint, tokens[domain])
                    # notify challenge are met
                    log("Starting key authorization")
                    keyauthorization = "{0}.{1}".format(tokens[domain], account_thumbprint)
                    code, result = self._send_signed(challenges[domain]['uri'], header, {
                        "resource": "challenge",
                        "keyAuthorization": keyauthorization,
                    })
                    if code != 202:
                        raise ValueError("Error triggering challenge: {0} {1}".format(code, result))

                    # wait for challenge to be verified
                    while True:
                        try:
                            resp = tools.get_url(challenges[domain]['uri'])
                            challenge_status = json.loads(resp.read().decode('utf8'))
                        except IOError as e:
                            raise ValueError("Error checking challenge: {0} {1}".format(
                                e.code, json.loads(e.read().decode('utf8'))))
                        if challenge_status['status'] == "pending":
                            time.sleep(2)
                        elif challenge_status['status'] == "valid":
                            log("{0} verified!".format(domain))
                            break
                        else:
                            raise ValueError("{0} challenge did not pass: {1}".format(
                                domain, challenge_status))
                finally:
                    challenge_handlers[domain].stop_challenge(domain, account_thumbprint, tokens[domain])
        finally:
            # Destroy challenge handlers in reverse order to replay
            # any saved state information in the handlers correctly
            for domain in reversed(domains):
                try:
                    challenge_handlers[domain].destroy_challenge(domain, account_thumbprint, tokens[domain])
                except Exception as e:
                    log('Challenge destruction failed: {}'.format(e), error=True)

        # get the new certificate
        log("Signing certificate...")
        code, result = self._send_signed(self.ca + "/acme/new-cert", header, {
            "resource": "new-cert",
            "csr": tools.bytes_to_base64url(tools.convert_cert_to_der_bytes(csr)),
        })
        if code != 201:
            raise ValueError("Error signing certificate: {0} {1}".format(code, result))

        # return signed certificate!
        log("Certificate signed!")
        cert = tools.convert_der_bytes_to_cert(result)
        return cert, tools.download_issuer_ca(cert)

    # @brief function to revoke a certificate using ACME
    # @param crt certificate to revoke
    # @param reason (int) optional certificate revoke reason (see https://tools.ietf.org/html/rfc5280#section-5.3.1)
    def revoke_crt(self, crt, reason=None):
        header = self._prepare_header()
        payload = {"resource": "revoke-cert",
                   "certificate": tools.bytes_to_base64url(tools.convert_cert_to_der_bytes(crt))}
        if reason:
            payload['reason'] = int(reason)
        code, result = self._send_signed(self.ca + "/acme/revoke-cert", header, payload)
        if code < 400:
            log("Revocation successful")
        else:
            raise ValueError("Revocation failed: {}".format(result))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# acertmgr - acme api v2 functions (implements RFC8555)
# Copyright (c) Rudolf Mayerhofer, 2019.
# available under the ISC license, see LICENSE

import copy
import json
import re
import time
from datetime import datetime

from acertmgr import tools
from acertmgr.authority.acme import ACMEAuthority as AbstractACMEAuthority
from acertmgr.tools import log

# Maximum age for nonce values (Boulder invalidates them after some time, so we use a low value of 2 minutes here)
MAX_NONCE_AGE = 120


class ACMEAuthority(AbstractACMEAuthority):
    # @brief Init class with config
    # @param config Configuration data
    # @param key Account key data
    def __init__(self, config, key):
        AbstractACMEAuthority.__init__(self, config, key)
        # Initialize config vars
        self.ca = config['authority']
        self.tos_agreed = str(config.get('authority_tos_agreement')).lower() == 'true'
        contact_email = config.get('authority_contact_email')
        if contact_email is None:
            self.contact = None
        elif isinstance(contact_email, list):
            self.contact = ["mailto:{}".format(contact) for contact in contact_email]
        else:
            self.contact = ["mailto:{}".format(contact_email)]

        # Initialize runtime vars
        code, self.directory, _ = self._request_url(self.ca + '/directory')
        if code >= 400 or not self.directory:
            self.directory = {
                "meta": {},
                "newAccount": "{}/acme/new-acct".format(self.ca),
                "newNonce": "{}/acme/new-nonce".format(self.ca),
                "newOrder": "{}/acme/new-order".format(self.ca),
                "revokeCert": "{}/acme/revoke-cert".format(self.ca),
            }
            log("API directory retrieval failed ({}). Guessed necessary values: {}".format(code, self.directory),
                warning=True)
        self.nonce = None
        self.nonce_time = 0

        self.algorithm, jwk = tools.get_key_alg_and_jwk(key)
        self.account_protected = {
            "alg": self.algorithm,
            "jwk": jwk
        }
        self.account_id = None  # will be updated to correct value during account registration

    # @brief fetch a given url
    def _request_url(self, url, data=None, raw_result=False):
        header = {'Content-Type': 'application/jose+json'}
        if data:
            # Always encode data to bytes
            data = data.encode('utf-8')
        try:
            resp = tools.get_url(url, data, header)
        except IOError as e:
            body = getattr(e, "read", e.__str__)()
            if getattr(body, 'decode', None):
                # Decode function available? Use it to get a proper str
                body = body.decode('utf-8')
            return getattr(e, "code", 999), body, {}

        # Store next Replay-Nonce if it is in the header
        if 'Replay-Nonce' in resp.headers:
            self.nonce = resp.headers['Replay-Nonce']
            self.nonce_time = time.time()

        body = resp.read()
        if getattr(body, 'decode', None):
            # Decode function available? Use it to get a proper str
            body = body.decode('utf-8')
        if not raw_result and len(body) > 0:
            try:
                body = json.loads(body)
            except json.JSONDecodeError as e:
                raise ValueError('Could not parse non-raw result (expected JSON)', e)

        return resp.getcode(), body, resp.headers

    # @brief fetch an url with a signed request
    def _request_acme_url(self, url, payload=None, protected=None, raw_result=False):
        if not protected:
            protected = {}

        if payload:
            payload64 = tools.bytes_to_base64url(json.dumps(payload).encode('utf8'))
        else:
            payload64 = ""  # for POST-as-GET

        # Request a new nonce if there is none in cache
        if not self.nonce or time.time() > self.nonce_time + MAX_NONCE_AGE:
            self._request_url(self.directory['newNonce'])
        # Set request nonce to current cache value
        protected["nonce"] = self.nonce
        # Reset nonce cache as we are using it's current value
        self.nonce = None

        protected["url"] = url
        if self.algorithm:
            protected["alg"] = self.algorithm
        if self.account_id:
            protected["kid"] = self.account_id
        protected64 = tools.bytes_to_base64url(json.dumps(protected).encode('utf8'))
        out = tools.signature_of_str(self.key, '.'.join([protected64, payload64]))
        data = json.dumps({
            "protected": protected64,
            "payload": payload64,
            "signature": tools.bytes_to_base64url(out),
        })
        return self._request_url(url, data, raw_result)

    # @brief send a signed request to authority
    def _request_acme_endpoint(self, request, payload=None, protected=None, raw_result=False):
        return self._request_acme_url(self.directory[request], payload, protected, raw_result)

    # @brief check if a given cert_profile is valid for this ACMEAuthority
    def _is_valid_cert_profile(self, profile):
        return (profile is not None
                and 'meta' in self.directory
                and 'profiles' in self.directory['meta']
                and profile in self.directory['meta']['profiles'])

    # @brief register an account over ACME
    def register_account(self):
        if self.account_id:
            # We already have registered with this authority, just return
            return

        protected = copy.deepcopy(self.account_protected)
        payload = {
            "termsOfServiceAgreed": self.tos_agreed,
            "onlyReturnExisting": False,
        }
        if self.contact:
            payload["contact"] = self.contact
        code, result, headers = self._request_acme_endpoint("newAccount", payload, protected)
        if code < 400 and result['status'] == 'valid':
            self.account_id = headers['Location']
            if 'meta' in self.directory and 'termsOfService' in self.directory['meta']:
                log("ToS at {} have been accepted.".format(self.directory['meta']['termsOfService']))
            log("Account registered and valid on {}.".format(self.ca))
        else:
            raise ValueError("Error registering account: {0} {1}".format(code, result))

    # @brief function to fetch certificate using ACME
    # @param csr the certificate signing request in pyopenssl format
    # @param domains list of domains in the certificate, first is CN
    # @param challenge_handlers a dict containing challenge for all given domains
    # @return the certificate and corresponding ca as a tuple
    # @note algorithm and parts of the code are from acme-tiny
    def get_crt_from_csr(self, csr, domains, challenge_handlers, cert_profile=None):
        account_thumbprint = tools.bytes_to_base64url(
            tools.hash_of_str(json.dumps(self.account_protected['jwk'], sort_keys=True, separators=(',', ':'))))

        log("Ordering certificate for {}".format(domains))
        identifiers = [{'type': 'dns', 'value': domain} for domain in domains]
        payload = {'identifiers': identifiers}
        if cert_profile is not None:
            if self._is_valid_cert_profile(cert_profile):
                payload['profile'] = cert_profile
            else:
                log("Ignoring invalid certificate profile for this authority: {}".format(cert_profile))
        code, order, headers = self._request_acme_endpoint('newOrder', payload)
        if code >= 400:
            raise ValueError("Error with certificate order: {0} {1}".format(code, order))

        order_url = headers['Location']
        authorizations = list()
        # verify each domain
        try:
            for authorizationUrl in order['authorizations']:
                # get new challenge
                code, authorization, _ = self._request_acme_url(authorizationUrl)
                if code >= 400:
                    raise ValueError("Error requesting authorization: {0} {1}".format(code, authorization))

                authorization['_domain'] = "*.{}".format(authorization['identifier']['value']) if \
                    'wildcard' in authorization and authorization['wildcard'] else authorization['identifier']['value']

                if authorization.get('status', 'no-status-found') == 'valid':
                    log("{} has already been authorized".format(authorization['_domain']))
                    continue
                if authorization['_domain'] not in challenge_handlers:
                    raise ValueError("No challenge handler given for domain: {0}".format(authorization['_domain']))
                log("Authorizing {0}".format(authorization['_domain']))

                # create the challenge
                ctype = challenge_handlers[authorization['_domain']].get_challenge_type()
                matching_challenges = [c for c in authorization['challenges'] if c['type'] == ctype]
                if len(matching_challenges) == 0:
                    raise ValueError("Error no challenge matching {0} found: {1}".format(ctype, authorization))

                authorization['_challenge'] = matching_challenges[0]
                if authorization['_challenge'].get('status', 'no-status-found') == 'valid':
                    log("{} has already been authorized using {}".format(authorization['_domain'], ctype))
                    continue

                authorization['_token'] = re.sub(r"[^A-Za-z0-9_\-]", "_", authorization['_challenge']['token'])
                challenge_handlers[authorization['_domain']].create_challenge(authorization['identifier']['value'],
                                                                              account_thumbprint,
                                                                              authorization['_token'])
                authorizations.append(authorization)

            # after all challenges are created, start processing authorizations
            for authorization in authorizations:
                try:
                    log("Starting verification of {}".format(authorization['_domain']))
                    challenge_handlers[authorization['_domain']].start_challenge(authorization['identifier']['value'],
                                                                                 account_thumbprint,
                                                                                 authorization['_token'])
                    # notify challenge is met
                    code, challenge_status, _ = self._request_acme_url(authorization['_challenge']['url'], {
                        "keyAuthorization": "{0}.{1}".format(authorization['_token'], account_thumbprint),
                    })
                    # wait for challenge to be verified
                    while code < 400 and challenge_status.get('status') == "pending":
                        time.sleep(5)
                        code, challenge_status, _ = self._request_acme_url(authorization['_challenge']['url'])

                    if code < 400 and challenge_status.get('status') == "valid":
                        log("{0} verified".format(authorization['_domain']))
                    else:
                        raise ValueError("{0} challenge did not pass ({1}): {2}".format(
                            authorization['_domain'], code, challenge_status))
                finally:
                    challenge_handlers[authorization['_domain']].stop_challenge(authorization['identifier']['value'],
                                                                                account_thumbprint,
                                                                                authorization['_token'])
        finally:
            # Destroy challenge handlers in reverse order to replay
            # any saved state information in the handlers correctly
            for authorization in reversed(authorizations):
                try:
                    challenge_handlers[authorization['_domain']].destroy_challenge(
                        authorization['identifier']['value'], account_thumbprint, authorization['_token'])
                except Exception as e:
                    log('Challenge destruction failed: {}'.format(e), error=True)

        # check order status and retry once
        code, order, _ = self._request_acme_url(order_url)
        if code < 400 and order.get('status') == 'pending':
            time.sleep(5)
            code, order, _ = self._request_acme_url(order_url)
        if code >= 400:
            raise ValueError("Order is still not ready to be finalized: {0} {1}".format(code, order))

        # get the new certificate
        log("Finalizing certificate")
        code, finalize, _ = self._request_acme_url(order['finalize'], {
            "csr": tools.bytes_to_base64url(tools.convert_cert_to_der_bytes(csr)),
        })
        while code < 400 and (finalize.get('status') == 'pending' or finalize.get('status') == 'processing'):
            time.sleep(5)
            code, finalize, _ = self._request_acme_url(order_url)
        if code >= 400:
            raise ValueError("Error finalizing certificate: {0} {1}".format(code, finalize))
        log("Certificate ready!")

        # return certificate
        code, certificate, _ = self._request_acme_url(finalize['certificate'], raw_result=True)
        if code >= 400:
            raise ValueError("Error downloading certificate chain: {0} {1}".format(code, certificate))

        cert_dict = re.match((r'(?P<cert>^-----BEGIN CERTIFICATE-----\n[^\-]+\n-----END CERTIFICATE-----)\n*'
                              r'(?P<ca>-----BEGIN CERTIFICATE-----\n.+\n-----END CERTIFICATE-----)?$'),
                             certificate, re.DOTALL).groupdict()
        cert = tools.convert_pem_str_to_cert(cert_dict['cert'])
        if cert_dict['ca'] is None:
            ca = tools.download_issuer_ca(cert)
        else:
            ca = tools.convert_pem_str_to_cert(cert_dict['ca'])

        return cert, ca

    # @brief function to revoke a certificate using ACME
    # @param crt certificate to revoke
    # @param reason (int) optional certificate revoke reason (see https://tools.ietf.org/html/rfc5280#section-5.3.1)
    def revoke_crt(self, crt, reason=None):
        payload = {'certificate': tools.bytes_to_base64url(tools.convert_cert_to_der_bytes(crt))}
        if reason:
            payload['reason'] = int(reason)
        code, result, _ = self._request_acme_endpoint("revokeCert", payload)
        if code < 400:
            log("Revocation successful")
        else:
            raise ValueError("Revocation failed: {}".format(result))

    # @brief check if the given certificate should be renewed according to it's RFC9773 data
    # @param crt certificate to check
    # @param issuer certificate necessary for correct AKI data
    # @return True if the certificate should be renewed, False otherwise
    def check_ari_for_renewal(self, crt, issuer):
        if 'renewalInfo' not in self.directory:
            return False
        try:
            sn = tools.get_cert_serialnumber_bytes(crt)
            aki = tools.get_issuer_cert_aki_bytes(issuer)
            url = "{}/{}.{}".format(self.directory['renewalInfo'],
                                    tools.bytes_to_base64url(aki),
                                    tools.bytes_to_base64url(sn))
            code, result, _ = self._request_url(url)
            if code >= 400:
                return False # Do not mark for renewal on error response
            start = datetime.fromisoformat(result['suggestedWindow']['start'])
            return datetime.now().timestamp() >= start.timestamp()  # True if window is reached or exceeded
        except Exception as e:
            log("Failed to check ARI for {} due to exception: {} - {}".format(tools.get_cert_identifier(crt),
                type(e), e), e)
            return False  # Do not mark for renewal in case anything has errored here



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# run as a module using "python -m acertmgr"
# Copyright (c) Rudolf Mayerhofer, 2019.
# available under the ISC license, see LICENSE

import acertmgr

if __name__ == "__main__":
    acertmgr.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# web - http based challenge handler
# Copyright (c) Rudolf Mayerhofer, 2019.
# available under the ISC license, see LICENSE

import os

from acertmgr import tools
from acertmgr.modes.abstract import AbstractChallengeHandler


class HTTPChallengeHandler(AbstractChallengeHandler):
    @staticmethod
    def get_challenge_type():
        return "http-01"

    def __init__(self, config):
        AbstractChallengeHandler.__init__(self, config)
        self.http_verify = str(config.get("http_verify", "true")).lower() == "true"

    def create_challenge(self, domain, thumbprint, token):
        raise NotImplementedError

    def destroy_challenge(self, domain, thumbprint, token):
        raise NotImplementedError

    def start_challenge(self, domain, thumbprint, token):
        if self.http_verify:
            keyauthorization = "{0}.{1}".format(token, thumbprint)
            wellknown_url = "http://{0}/.well-known/acme-challenge/{1}".format(domain, token)
            try:
                resp = tools.get_url(wellknown_url)
                resp_data = resp.read().decode('utf8').strip()
                if resp_data != keyauthorization:
                    raise ValueError("keyauthorization and response data do NOT match")
            except (IOError, ValueError):
                raise ValueError("keyauthorization verification failed")


class ChallengeHandler(HTTPChallengeHandler):
    def __init__(self, config):
        HTTPChallengeHandler.__init__(self, config)
        self.challenge_directory = config.get("webdir", "/var/www/acme-challenge/")
        if not os.path.isdir(self.challenge_directory):
            raise FileNotFoundError("Challenge directory (%s) does not exist!" % self.challenge_directory)

    def create_challenge(self, domain, thumbprint, token):
        keyauthorization = "{0}.{1}".format(token, thumbprint)
        wellknown_path = os.path.join(self.challenge_directory, token)
        with open(wellknown_path, "w") as wellknown_file:
            wellknown_file.write(keyauthorization)

    def destroy_challenge(self, domain, thumbprint, token):
        os.remove(os.path.join(self.challenge_directory, token))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# config - acertmgr config parser
# Copyright (c) Markus Hauschild & David Klaftenegger, 2016.
# Copyright (c) Rudolf Mayerhofer, 2019.
# available under the ISC license, see LICENSE

import argparse
import copy
import hashlib
import io
import json
import os

from acertmgr.tools import idna_convert

# Configuration defaults to use if not specified otherwise
DEFAULT_CONF_DIR = "/etc/acertmgr"
DEFAULT_CONF_FILENAME = "acertmgr.conf"
DEFAULT_TTL = 0.33333  # percent of the certificates total lifetime
DEFAULT_VALIDATE_ARI = "true"
DEFAULT_VALIDATE_OCSP = "sha1" # mandated by RFC5019
DEFAULT_API = "v2"
DEFAULT_AUTHORITY = "https://acme-v02.api.letsencrypt.org"


# @brief augment configuration with defaults
# @param domainconfig the domain configuration
# @param defaults the default configuration
# @return the augmented configuration
def complete_action_config(domainconfig, config):
    defaults = config['defaults']
    domainconfig['ca_file'] = config['ca_file']
    domainconfig['cert_file'] = config['cert_file']
    domainconfig['key_file'] = config['key_file']
    for name, value in defaults.items():
        if name not in domainconfig:
            domainconfig[name] = value
    if 'action' not in domainconfig:
        domainconfig['action'] = None
    return domainconfig


# @brief update config[name] with value from localconfig>globalconfig>default
def update_config_value(config, name, localconfig, globalconfig, default):
    values = [x[name] for x in localconfig if name in x]
    if len(values) > 0:
        config[name] = values[0]
    else:
        config[name] = globalconfig.get(name, default)


# @brief parse authority from config
def parse_authority(localconfig, globalconfig, runtimeconfig):
    authority = {}
    # - API version
    update_config_value(authority, 'api', localconfig, globalconfig, DEFAULT_API)

    # - Certificate authority
    update_config_value(authority, 'authority', localconfig, globalconfig, DEFAULT_AUTHORITY)

    # - Certificate authority ToS agreement
    update_config_value(authority, 'authority_tos_agreement', localconfig, globalconfig,
                        runtimeconfig['authority_tos_agreement'])

    # - Certificate authority contact email addresses
    update_config_value(authority, 'authority_contact_email', localconfig, globalconfig, None)

    # - Account key path
    update_config_value(authority, 'account_key', localconfig, globalconfig,
                        os.path.join(runtimeconfig['work_dir'], "account.key"))

    # - Account key algorithm (if key has to be (re-)generated)
    update_config_value(authority, 'account_key_algorithm', localconfig, globalconfig, None)

    # - Account key length (if key has to be (re-)generated, converted to int)
    update_config_value(authority, 'account_key_length', localconfig, globalconfig, None)
    authority['account_key_length'] = int(authority['account_key_length']) if authority['account_key_length'] else None

    return authority


# @brief load the configuration from a file
def parse_config_entry(entry, globalconfig, runtimeconfig):
    config = dict()

    # Basic domain information
    domains, localconfig = entry
    config['domainlist'] = domains.split(' ')
    config['id'] = hashlib.md5(domains.encode('utf-8')).hexdigest()

    # Convert unicode to IDNA domains
    config['domainlist_idna_mapped'] = {}
    for idx in range(0, len(config['domainlist'])):
        if any(ord(c) >= 128 for c in config['domainlist'][idx]):
            domain_human = config['domainlist'][idx]
            domain_idna = idna_convert(domain_human)
            if domain_idna != domain_human:
                config['domainlist'][idx] = domain_idna  # Update domain with idna counterpart
                config['domainlist_idna_mapped'][domain_idna] = domain_human  # Store original domain for reference

    # Action config defaults
    config['defaults'] = globalconfig.get('defaults', {})

    # Authority related config options
    config['authority'] = parse_authority(localconfig, globalconfig, runtimeconfig)

    # Certificate directory
    update_config_value(config, 'cert_dir', localconfig, globalconfig, runtimeconfig['work_dir'])

    # TTL days
    update_config_value(config, 'ttl_days', localconfig, globalconfig, DEFAULT_TTL)
    config['ttl_days'] = float(config['ttl_days'])

    # Validate ARI on certificate verification
    update_config_value(config, 'validate_ari', localconfig, globalconfig, DEFAULT_VALIDATE_ARI)

    # Validate OCSP on certificate verification
    update_config_value(config, 'validate_ocsp', localconfig, globalconfig, DEFAULT_VALIDATE_OCSP)

    # Revoke old certificate with reason superseded after renewal
    update_config_value(config, 'cert_revoke_superseded', localconfig, globalconfig, "false")

    # Whether to include request for OCSP must-staple in the certificate
    update_config_value(config, 'cert_must_staple', localconfig, globalconfig, "false")

    # Use a specific certificate profile that the authority provides
    update_config_value(config, 'cert_profile', localconfig, globalconfig, None)

    # Use a static cert request
    update_config_value(config, 'csr_static', localconfig, globalconfig, "false")

    # SSL key algorithm (if key has to be (re-)generated)
    update_config_value(config, 'key_algorithm', localconfig, globalconfig, None)
    # Update config id if we have a key algorithm set to allow for
    # multiple certs with different algorithms for the same set of domains
    if config.get('key_algorithm', None):
        config['id'] += "_" + config['key_algorithm'].lower()

    # SSL key length (if key has to be (re-)generated, converted to int)
    update_config_value(config, 'key_length', localconfig, globalconfig, None)
    config['key_length'] = int(config['key_length']) if config['key_length'] else None

    # SSL cert request location
    update_config_value(config, 'csr_file', localconfig, globalconfig,
                        os.path.join(config['cert_dir'], "{}.csr".format(config['id'])))

    # SSL cert location (with compatibility to older versions)
    update_config_value(config, 'cert_file', localconfig, globalconfig,
                        os.path.join(config['cert_dir'], "{}.crt".format(config['id'])))

    # SSL key location (with compatibility to older versions)
    update_config_value(config, 'key_file', localconfig, globalconfig,
                        os.path.join(config['cert_dir'], "{}.key".format(config['id'])))

    # SSL CA location / use static
    update_config_value(config, 'ca_file', localconfig, globalconfig,
                        os.path.join(config['cert_dir'], "{}.ca".format(config['id'])))
    update_config_value(config, 'ca_static', localconfig, globalconfig, "false")

    # Domain action configuration
    config['actions'] = list()
    for actioncfg in [x for x in localconfig if 'path' in x]:
        config['actions'].append(complete_action_config(actioncfg, config))

    # Domain challenge handler configuration
    config['handlers'] = dict()
    handlerconfigs = [x for x in localconfig if 'mode' in x]
    _domaintranslation_dict = {x: y for x, y in config.get('domaintranslation', [])}
    for domain in config['domainlist']:
        # Use global config as base handler config
        cfg = copy.deepcopy(globalconfig)

        # Determine generic domain handler config values
        genericfgs = [x for x in handlerconfigs if 'domain' not in x]
        if len(genericfgs) > 0:
            cfg.update(genericfgs[0])

        # Update handler config with more specific values (use original names for translated unicode domains)
        specificcfgs = [x for x in handlerconfigs if
                        'domain' in x and x['domain'] == config['domainlist_idna_mapped'].get(domain, domain)]
        if len(specificcfgs) > 0:
            cfg.update(specificcfgs[0])

        config['handlers'][domain] = cfg

    return config


# @brief load the configuration from a file
def load():
    runtimeconfig = dict()
    parser = argparse.ArgumentParser(description="acertmgr - Automated Certificate Manager using ACME/Let's Encrypt")
    parser.add_argument("-c", "--config-file", nargs="?",
                        help="global configuration file (default='$config_dir/{}')".format(DEFAULT_CONF_FILENAME))
    parser.add_argument("-d", "--config-dir", nargs="?",
                        help="domain configuration directory (default='{}')".format(DEFAULT_CONF_DIR))
    parser.add_argument("-w", "--work-dir", nargs="?",
                        help="persistent work data directory (default='$config_dir')")
    parser.add_argument("--authority-tos-agreement", "--tos-agreement", "--tos", nargs="?",
                        help="Agree to the authorities Terms of Service (value required depends on authority)")
    parser.add_argument("--force-renew", "--renew-now", nargs="?",
                        help="Renew all domain configurations matching the given value immediately")
    parser.add_argument("--revoke", nargs="?",
                        help="Revoke a certificate file issued with the currently configured account key.")
    parser.add_argument("--revoke-reason", nargs="?", type=int,
                        help="Provide a revoke reason, see https://tools.ietf.org/html/rfc5280#section-5.3.1")
    args = parser.parse_args()

    # Determine domain configuration directory
    if args.config_dir:
        domain_config_dir = args.config_dir
    else:
        domain_config_dir = DEFAULT_CONF_DIR

    # Determine global configuration file
    if args.config_file:
        global_config_file = args.config_file
    else:
        global_config_file = os.path.join(domain_config_dir, DEFAULT_CONF_FILENAME)

    # Runtime configuration: Get from command-line options
    # - work_dir
    if args.work_dir:
        runtimeconfig['work_dir'] = args.work_dir
    else:
        runtimeconfig['work_dir'] = domain_config_dir
    #  create work_dir if it does not exist yet
    if not os.path.isdir(runtimeconfig['work_dir']):
        os.mkdir(runtimeconfig['work_dir'], int("0700", 8))

    # - authority_tos_agreement
    if args.authority_tos_agreement:
        runtimeconfig['authority_tos_agreement'] = args.authority_tos_agreement
    else:
        runtimeconfig['authority_tos_agreement'] = None

    # - force-rewew
    if args.force_renew:
        domaintranslation = [idna_convert(d) for d in args.force_renew.split(' ')]
        if len(domaintranslation) > 0:
            runtimeconfig['force_renew'] = domaintranslation
        else:
            runtimeconfig['force_renew'] = args.force_renew.split(' ')

    # - revoke
    if args.revoke:
        runtimeconfig['mode'] = 'revoke'
        runtimeconfig['revoke'] = args.revoke
        runtimeconfig['revoke_reason'] = args.revoke_reason

    # Global configuration: Load from file
    globalconfig = dict()
    if os.path.isfile(global_config_file):
        with io.open(global_config_file) as config_fd:
            try:
                globalconfig = json.load(config_fd)
            except ValueError:
                import yaml
                config_fd.seek(0)
                globalconfig = yaml.safe_load(config_fd)

    # Domain configuration(s): Load from file(s)
    domainconfigs = list()
    if os.path.isdir(domain_config_dir):
        for domain_config_file in os.listdir(domain_config_dir):
            domain_config_file = os.path.join(domain_config_dir, domain_config_file)
            # check file extension and skip if global config file
            if domain_config_file.endswith(".conf") and \
                    os.path.abspath(domain_config_file) != os.path.abspath(global_config_file):
                with io.open(domain_config_file) as config_fd:
                    try:
                        data = json.load(config_fd)
                    except ValueError:
                        import yaml
                        config_fd.seek(0)
                        data = yaml.safe_load(config_fd)
                    if isinstance(data, list):
                        # Handle newer config in list format (allows for multiple entries with same domains)
                        entries = list()
                        for element in data:
                            entries += element.items()
                    else:
                        # Handle older config format with just one entry per same domain set
                        entries = data.items()
                    for entry in entries:
                        domainconfigs.append(parse_config_entry(entry, globalconfig, runtimeconfig))

    # Define a fallback authority from global configuration / defaults
    runtimeconfig['fallback_authority'] = parse_authority([], globalconfig, runtimeconfig)

    return runtimeconfig, domainconfigs


if __name__ == '__main__':
    # Simple configuration load test and output
    from pprint import pprint
    pprint(load())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# dns.nsupdate - rfc2136 based challenge handler
# Copyright (c) Rudolf Mayerhofer, 2018-2019
# available under the ISC license, see LICENSE

import ipaddress
import re
import socket
import time
from datetime import datetime, timedelta

import dns
import dns.query
import dns.resolver
import dns.tsigkeyring
import dns.update

from acertmgr import tools
from acertmgr.modes.abstract import AbstractChallengeHandler
from acertmgr.tools import log

QUERY_TIMEOUT = 60  # seconds are the maximum for any query (otherwise the DNS server will be considered dead)
REGEX_IP4 = r'^(?:[0-9]{1,3}\.){3}[0-9]{1,3}$'
REGEX_IP6 = r'^(([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,7}' \
            r':|([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,5}' \
            r'(:[0-9a-fA-F]{1,4}){1,2}|([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}' \
            r'|([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|([0-9a-fA-F]{1,4}:){1,2}' \
            r'(:[0-9a-fA-F]{1,4}){1,5}|[0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})' \
            r'|:((:[0-9a-fA-F]{1,4}){1,7}|:)|fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}' \
            r'|::(ffff(:0{1,4}){0,1}:){0,1}((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}' \
            r'(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|([0-9a-fA-F]{1,4}:){1,4}' \
            r':((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9]))$'
_lookup_ip_cache = {}
_lookup_ns_ip_cache = {}
_lookup_zone_cache = {}


class DNSChallengeHandler(AbstractChallengeHandler):
    @staticmethod
    def _lookup_ip(domain_or_ip):
        if domain_or_ip in _lookup_ip_cache:
            return _lookup_ip_cache[domain_or_ip]

        try:
            if re.search(REGEX_IP4, domain_or_ip.strip()) or re.search(REGEX_IP6, domain_or_ip.strip()):
                return str(ipaddress.ip_address(domain_or_ip))
        except ValueError:
            pass
        # No valid ip found so far, try to resolve using system resolver
        result = socket.getaddrinfo(domain_or_ip, 53)
        if len(result) > 0:
            retval = result[0][4][0]
            _lookup_ip_cache[domain_or_ip] = retval
            return retval
        else:
            raise ValueError("Could not lookup dns ip for {}".format(domain_or_ip))

    @staticmethod
    def _lookup_ns_ip(domain, nameserver=None):
        zone, zonemaster = DNSChallengeHandler._lookup_zone(domain, nameserver)
        cache_key = "{}${}".format(zone, zonemaster)
        if cache_key in _lookup_ns_ip_cache:
            return _lookup_ns_ip_cache[cache_key]

        if not nameserver:
            nameserver = DNSChallengeHandler._lookup_ip(zonemaster)

        request = dns.message.make_query(zone, dns.rdatatype.NS)
        response = dns.query.udp(request, nameserver, timeout=QUERY_TIMEOUT)
        retval = set()
        if response.rcode() == dns.rcode.NOERROR:
            for answer in response.answer:
                for item in answer:
                    if item.rdtype == dns.rdatatype.NS:
                        retval.add(DNSChallengeHandler._lookup_ip(item.to_text()))
            _lookup_ns_ip_cache[cache_key] = retval
        return retval

    @staticmethod
    def _lookup_zone(domain, nameserver=None):
        cache_key = "{}${}".format(domain, nameserver)
        if cache_key in _lookup_zone_cache:
            return _lookup_zone_cache[cache_key]

        if nameserver:
            nameservers = [nameserver]
        else:
            nameservers = dns.resolver.get_default_resolver().nameservers

        domain = dns.name.from_text(domain)
        if not domain.is_absolute():
            domain = domain.concatenate(dns.name.root)

        while domain.parent() != dns.name.root:
            request = dns.message.make_query(domain, dns.rdatatype.SOA)
            for nameserver in nameservers:
                try:
                    response = dns.query.udp(request, nameserver, timeout=QUERY_TIMEOUT)
                    if response.rcode() == dns.rcode.NOERROR:
                        for answer in response.answer:
                            for item in answer:
                                if item.rdtype == dns.rdatatype.SOA:
                                    zone = domain.to_text()
                                    authoritative_ns = item.mname.to_text().split(' ')[0]
                                    retval = zone, authoritative_ns
                                    _lookup_zone_cache[cache_key] = retval
                                    return retval
                    else:
                        break
                except dns.exception.Timeout:
                    # Go to next nameserver on timeout
                    continue
                except dns.exception.DNSException:
                    # Break loop on any other error
                    break
            domain = domain.parent()
        raise ValueError('No zone SOA for "{0}"'.format(domain))

    @staticmethod
    def _check_txt_record_value(domain, txtvalue, nameserverip, use_tcp=False):
        try:
            request = dns.message.make_query(domain, dns.rdatatype.TXT)
            if use_tcp:
                response = dns.query.tcp(request, nameserverip, timeout=QUERY_TIMEOUT)
            else:
                response = dns.query.udp(request, nameserverip, timeout=QUERY_TIMEOUT)
            for rrset in response.answer:
                for answer in rrset:
                    if answer.to_text().strip('"') == txtvalue:
                        return True
        except dns.exception.DNSException:
            # Ignore DNS errors and return failure
            return False

    @staticmethod
    def _determine_txtvalue(thumbprint, token):
        return tools.bytes_to_base64url(tools.hash_of_str("{0}.{1}".format(token, thumbprint)))

    @staticmethod
    def get_challenge_type():
        return "dns-01"

    def __init__(self, config):
        AbstractChallengeHandler.__init__(self, config)
        self.dns_updatedomain = config.get("dns_updatedomain")
        self.dns_ttl = int(config.get("dns_ttl", 60))
        self.dns_verify_waittime = int(config.get("dns_verify_waittime", 2 * self.dns_ttl))
        self.dns_verify_failtime = int(config.get("dns_verify_failtime", self.dns_verify_waittime + 1))
        self.dns_verify_interval = int(config.get("dns_verify_interval", 10))
        self.dns_verify_all_ns = str(config.get("dns_verify_all_ns")).lower() == "true"
        self.dns_verify_server = config.get("dns_verify_server")

        self._valid_times = {}

    def _determine_challenge_domain(self, domain):
        if self.dns_updatedomain:
            domain = self.dns_updatedomain
        else:
            domain = "_acme-challenge.{0}".format(domain)

        domain = dns.name.from_text(domain)
        if not domain.is_absolute():
            domain = domain.concatenate(dns.name.root)

        return domain.to_text()

    def create_challenge(self, domain, thumbprint, token):
        domain = self._determine_challenge_domain(domain)
        txtvalue = self._determine_txtvalue(thumbprint, token)
        self.add_dns_record(domain, txtvalue)
        self._valid_times[domain] = datetime.now() + timedelta(seconds=self.dns_verify_waittime)

    def add_dns_record(self, domain, txtvalue):
        raise NotImplementedError

    def destroy_challenge(self, domain, thumbprint, token):
        domain = self._determine_challenge_domain(domain)
        txtvalue = self._determine_txtvalue(thumbprint, token)
        self.remove_dns_record(domain, txtvalue)

    def remove_dns_record(self, domain, txtvalue):
        raise NotImplementedError

    def start_challenge(self, domain, thumbprint, token):
        domain = self._determine_challenge_domain(domain)
        txtvalue = self._determine_txtvalue(thumbprint, token)
        failtime = datetime.now() + timedelta(seconds=self.dns_verify_failtime)
        if self.verify_dns_record(domain, txtvalue):
            return
        else:
            log("Waiting until TXT record '{}' is ready".format(domain))
            while failtime > datetime.now():
                time.sleep(self.dns_verify_interval)
                if self.verify_dns_record(domain, txtvalue):
                    return
            raise ValueError("DNS challenge is not ready after waiting {} seconds".format(self.dns_verify_waittime))

    def verify_dns_record(self, domain, txtvalue):
        if self.dns_verify_all_ns:
            try:
                nameserverip = None
                if self.dns_verify_server:
                    # Use the specific dns server to determine NS for domain, will otherwise default to SOA master
                    nameserverip = self._lookup_ip(self.dns_verify_server)
                ns_ip = self._lookup_ns_ip(domain, nameserverip)
                if len(ns_ip) > 0 and all(self._check_txt_record_value(domain, txtvalue, ip) for ip in ns_ip):
                    # All NS servers have the necessary TXT record. Succeed immediately!
                    log("All NS ({}) for '{}' have the correct TXT record".format(','.join(ns_ip), domain))
                    return True
            except (ValueError, dns.exception.DNSException):
                # Fall back to next verification
                pass

        if self.dns_verify_server and not self.dns_verify_all_ns:
            try:
                # Verify using specific dns server
                nameserverip = self._lookup_ip(self.dns_verify_server)
                if self._check_txt_record_value(domain, txtvalue, nameserverip):
                    # Verify server confirms the necessary TXT record. Succeed immediately!
                    log("DNS server '{}' found correct TXT record for '{}'".format(self.dns_verify_server, domain))
                    return True
            except (ValueError, dns.exception.DNSException):
                # Fall back to next verification
                pass

        if domain not in self._valid_times:
            # No valid wait time for domain. Verification fails!
            return False
        # Verification fails or succeeds based on valid wait time set by add_dns_record
        return datetime.now() >= self._valid_times[domain]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# acertmgr - various support functions
# Copyright (c) Markus Hauschild & David Klaftenegger, 2016.
# Copyright (c) Rudolf Mayerhofer, 2019.
# available under the ISC license, see LICENSE

import base64
import io
import math
import os
import re
import stat
import sys
import traceback
from datetime import datetime, timedelta, timezone

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa, ec, padding
from cryptography.hazmat.primitives.asymmetric.utils import decode_dss_signature
from cryptography.utils import int_to_bytes
from cryptography.x509.oid import NameOID, ExtensionOID

try:
    from cryptography.x509 import ocsp
except ImportError:
    pass

try:
    from cryptography.hazmat.primitives.asymmetric import ed25519, ed448
except ImportError:
    pass

try:
    from urllib.request import urlopen, Request  # Python 3
except ImportError:
    from urllib2 import urlopen, Request  # Python 2

LOG_REPLACEMENTS = {}


class InvalidCertificateError(Exception):
    pass


# @brief a simple, portable indent function
def indent(text, spaces=0):
    ind = ' ' * spaces
    return os.linesep.join(ind + line for line in text.splitlines())


# @brief wrapper for log output
def log(msg, exc=None, error=False, warning=False):
    if error:
        prefix = "Error: "
    elif warning:
        prefix = "Warning: "
    else:
        prefix = ""

    output = prefix + msg
    for k, v in LOG_REPLACEMENTS.items():
        output = output.replace(k, v)

    if exc:
        _, exc_value, _ = sys.exc_info()
        if not getattr(exc, '__traceback__', None) and exc == exc_value:
            # Traceback handling on Python 2 is ugly, so we only output it if the exception is the current sys one
            formatted_exc = traceback.format_exc()
        else:
            formatted_exc = traceback.format_exception(type(exc), exc, getattr(exc, '__traceback__', None))
        exc_string = ''.join(formatted_exc) if isinstance(formatted_exc, list) else str(formatted_exc)
        output += os.linesep + indent(exc_string, len(prefix))

    if error or warning:
        sys.stderr.write(output + os.linesep)
        sys.stderr.flush()  # force flush buffers after message was written for immediate display
    else:
        sys.stdout.write(output + os.linesep)
        sys.stdout.flush()  # force flush buffers after message was written for immediate display


# @brief wrapper for downloading an url
def get_url(url, data=None, headers=None):
    return urlopen(Request(url, data=data, headers={} if headers is None else headers))

# @brief get cryptography version stable valid datetime objects
# @param crt_file string containing the path to the certificate file
def _cert_validity_utc_dates(cert):
    try:
        # cryptography >= 42: timezone-aware
        nv_before = cert.not_valid_before_utc
        nv_after = cert.not_valid_after_utc
    except AttributeError:
        # Older cryptography: properties are naive UTC datetimes
        nv_before = cert.not_valid_before
        nv_after = cert.not_valid_after
        if nv_before.tzinfo is None:
            nv_before = nv_before.replace(tzinfo=timezone.utc)
        if nv_after.tzinfo is None:
            nv_after = nv_after.replace(tzinfo=timezone.utc)
    return nv_before, nv_after

# @brief check whether existing certificate is still valid or expiring soon
# @param crt_file string containing the path to the certificate file
# @param ttl_days the minimum amount of days for which the certificate must be valid 
#        or if between 0 and 1 a fraction of the certificates total lifetime
# @return True if certificate is still valid for at least ttl_days, False otherwise
def is_cert_valid(cert, ttl_days):
    now = datetime.now(timezone.utc)
    nv_before, nv_after = _cert_validity_utc_dates(cert)
    if nv_before > now:
        raise InvalidCertificateError("Certificate seems to be from the future")
    elif ttl_days > 0 and ttl_days < 1:
        # If ttl_days is between 0 and 1 then it denominates a fraction of the total cert lifetime
        cert_lifetime = math.ceil((nv_after - nv_before).total_seconds() / 86400)
        # Set ttl_days to the real value in days based on certificate lifetime rounded up
        ttl_days = float(cert_lifetime) * ttl_days
    return (nv_after - now) > timedelta(days=ttl_days)


# @brief create a certificate signing request
# @param names list of domain names the certificate should be valid for
# @param key the key to use with the certificate in pyopenssl format
# @param must_staple whether or not the certificate should include the OCSP must-staple flag
# @return the CSR in pyopenssl format
def new_cert_request(names, key, must_staple=False):
    primary_name = x509.Name([x509.NameAttribute(NameOID.COMMON_NAME,
                                                 names[0].decode('utf-8') if getattr(names[0], 'decode', None) else
                                                 names[0])])
    all_names = x509.SubjectAlternativeName(
        [x509.DNSName(name.decode('utf-8') if getattr(name, 'decode', None) else name) for name in names])
    req = x509.CertificateSigningRequestBuilder()
    req = req.subject_name(primary_name)
    req = req.add_extension(all_names, critical=False)
    if must_staple:
        if getattr(x509, 'TLSFeature', None):
            req = req.add_extension(x509.TLSFeature(features=[x509.TLSFeatureType.status_request]), critical=False)
        else:
            log('OCSP must-staple ignored as current version of cryptography does not support the flag.', warning=True)
    req = req.sign(key, hashes.SHA256(), default_backend())
    return req


# @brief generate a new account key
# @param path path where the new key file should be written in PEM format (optional)
def new_account_key(path=None, key_algo=None, key_size=None):
    return new_ssl_key(path, key_algo, key_size)


# @brief generate a new ssl key
# @param path path where the new key file should be written in PEM format (optional)
def new_ssl_key(path=None, key_algo=None, key_size=None):
    if not key_algo or key_algo.lower() == 'rsa':
        if not key_size:
            key_size = 4096
        key_format = serialization.PrivateFormat.TraditionalOpenSSL
        private_key = rsa.generate_private_key(
            public_exponent=65537,
            key_size=key_size,
            backend=default_backend()
        )
    elif key_algo.lower() == 'ec' or key_algo.lower() == 'ecc':
        if not key_size or key_size == 256:
            key_curve = ec.SECP256R1
        elif key_size == 384:
            key_curve = ec.SECP384R1
        elif key_size == 521:
            key_curve = ec.SECP521R1
        else:
            raise ValueError("Unsupported EC curve size parameter: {}".format(key_size))
        key_format = serialization.PrivateFormat.PKCS8
        private_key = ec.generate_private_key(curve=key_curve, backend=default_backend())
    elif key_algo.lower() == 'ed25519' and "cryptography.hazmat.primitives.asymmetric.ed25519":
        key_format = serialization.PrivateFormat.PKCS8
        private_key = ed25519.Ed25519PrivateKey.generate()
    elif key_algo.lower() == 'ed448' and "cryptography.hazmat.primitives.asymmetric.ed448":
        key_format = serialization.PrivateFormat.PKCS8
        private_key = ed448.Ed448PrivateKey.generate()
    else:
        raise ValueError("Unsupported key algorithm: {}".format(key_algo))
    if path is not None:
        pem = private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=key_format,
            encryption_algorithm=serialization.NoEncryption(),
        )
        with io.open(path, 'wb') as pem_out:
            pem_out.write(pem)
        if hasattr(os, 'chmod'):
            try:
                os.chmod(path, int("0400", 8))
            except OSError:
                log('Could not set file permissions on {0}!'.format(path), warning=True)
        else:
            log('Keyfile permission handling unavailable on this platform', warning=True)
    return private_key


# @brief read a key from file
# @param path path to file
# @param key indicate whether we are loading a key
# @param csr indicate whether we are loading a csr
# @return the key in pyopenssl format
def read_pem_file(path, key=False, csr=False):
    with io.open(path, 'r') as f:
        if key:
            return serialization.load_pem_private_key(f.read().encode('utf-8'), None, default_backend())
        elif csr:
            return x509.load_pem_x509_csr(f.read().encode('utf8'), default_backend())
        else:
            return convert_pem_str_to_cert(f.read())


# @brief write cert data to PEM formatted file
def write_pem_file(crt, path, perms=None):
    if hasattr(os, 'chmod') and os.path.exists(path):
        try:
            os.chmod(path, os.stat(path).st_mode | stat.S_IWRITE)
        except OSError:
            log('Could not make file ({0}) writable'.format(path), warning=True)
    with io.open(path, "w") as f:
        f.write(convert_cert_to_pem_str(crt))
    if perms:
        if hasattr(os, 'chmod'):
            try:
                os.chmod(path, perms)
            except OSError:
                log('Could not set file permissions ({0}) on {1}!'.format(perms, path), warning=True)
        else:
            log('PEM-File permission handling unavailable on this platform', warning=True)


# @brief download the issuer ca for a given certificate
# @param cert certificate data
# @returns ca certificate data
def download_issuer_ca(cert):
    aia = cert.extensions.get_extension_for_oid(ExtensionOID.AUTHORITY_INFORMATION_ACCESS)
    ca_issuers = None
    for data in aia.value:
        if data.access_method == x509.OID_CA_ISSUERS:
            ca_issuers = data.access_location.value
            break

    if not ca_issuers:
        log("Could not determine issuer CA for given certificate: {}".format(cert), error=True)
        return None

    log("Downloading CA certificate from {}".format(ca_issuers))
    resp = get_url(ca_issuers)
    code = resp.getcode()
    if code >= 400:
        log("Could not download issuer CA (error {}) for given certificate: {}".format(code, cert), error=True)
        return None

    return x509.load_der_x509_certificate(resp.read(), default_backend())


# @brief determine all san domains on a given certificate
def get_cert_domains(cert):
    san_cert = cert.extensions.get_extension_for_oid(ExtensionOID.SUBJECT_ALTERNATIVE_NAME)
    domains = set()
    cns = cert.subject.get_attributes_for_oid(NameOID.COMMON_NAME)
    if len(cns) > 0:
        domains.add(cns[0].value)
    if san_cert:
        for d in san_cert.value:
            domains.add(d.value)
    return domains


# @brief determine certificate cn
def get_cert_identifier(cert):
    # Return the first domain as the certificates identifier (usually the CN, see above)
    return "{}".format(list(get_cert_domains(cert))[0])


# @brief determine certificate end of validity
def get_cert_valid_until(cert):
    _, nv_after = _cert_validity_utc_dates(cert)
    return nv_after


# @brief determine certificate serial number as bytes
def get_cert_serialnumber_bytes(cert):
    serial = cert.serial_number
    byte_count = (serial.bit_length() + 7) // 8  # Round up to nearest byte
    return serial.to_bytes(byte_count)


# @brief determine issuer certificate AKI bytes
def get_issuer_cert_aki_bytes(issuer_cert):
    if isinstance(issuer_cert, list):
        # Take the first certificate from the list as the issuer
        issuer_cert = issuer_cert[0]
    # Determine necessary AKI bytes
    ski = issuer_cert.extensions.get_extension_for_class(x509.SubjectKeyIdentifier)
    return x509.AuthorityKeyIdentifier.from_issuer_subject_key_identifier(ski.value).key_identifier

# @brief convert certificate to PEM format
# @param cert certificate object or a list thereof
# @return the certificate in PEM format
def convert_cert_to_pem_str(cert):
    if not isinstance(cert, list):
        cert = [cert]
    result = list()
    for data in cert:
        result.append(data.public_bytes(serialization.Encoding.PEM).decode('utf8'))
    return '\n'.join(result)


# @brief load a PEM certificate from str
# @return a certificate object or a list of objects if multiple are in the string
def convert_pem_str_to_cert(certdata):
    certs = re.findall(r'(-----BEGIN CERTIFICATE-----\n[^\-]+\n-----END CERTIFICATE-----)',
                       certdata, re.DOTALL)
    result = list()
    for data in certs:
        result.append(x509.load_pem_x509_certificate(data.encode('utf8'), default_backend()))
    return result[0] if len(result) == 1 else result


# @brief serialize cert/csr to DER bytes
def convert_cert_to_der_bytes(data):
    return data.public_bytes(serialization.Encoding.DER)


# @brief load a DER certificate from str
def convert_der_bytes_to_cert(data):
    return x509.load_der_x509_certificate(data, default_backend())


# @brief determine key signing algorithm and jwk data
# @return key algorithm, signature algorithm, key numbers as a dict
def get_key_alg_and_jwk(key):
    if isinstance(key, rsa.RSAPrivateKey):
        # See https://tools.ietf.org/html/rfc7518#section-6.3
        numbers = key.public_key().public_numbers()
        return "RS256", {"kty": "RSA",
                         "e": bytes_to_base64url(int_to_bytes(numbers.e)),
                         "n": bytes_to_base64url(int_to_bytes(numbers.n))}
    elif isinstance(key, ec.EllipticCurvePrivateKey):
        # See https://tools.ietf.org/html/rfc7518#section-6.2
        numbers = key.public_key().public_numbers()
        if isinstance(numbers.curve, ec.SECP256R1):
            alg = 'ES256'
            crv = 'P-256'
        elif isinstance(numbers.curve, ec.SECP384R1):
            alg = 'ES384'
            crv = 'P-384'
        elif isinstance(numbers.curve, ec.SECP521R1):
            alg = 'ES512'
            crv = 'P-521'
        else:
            raise ValueError("Unsupported EC curve in key: {}".format(key))
        full_octets = (int(crv[2:]) + 7) // 8
        return alg, {"kty": "EC", "crv": crv,
                     "x": bytes_to_base64url(int_to_bytes(numbers.x, full_octets)),
                     "y": bytes_to_base64url(int_to_bytes(numbers.y, full_octets))}
    elif "cryptography.hazmat.primitives.asymmetric.ed25519" in sys.modules and isinstance(key,
                                                                                           ed25519.Ed25519PrivateKey):
        # See https://tools.ietf.org/html/rfc8037#appendix-A.2
        return "EdDSA", {"kty": "OKP", "crv": "Ed25519",
                         "x": bytes_to_base64url(key.public_key().public_bytes(encoding=serialization.Encoding.Raw,
                                                                               format=serialization.PublicFormat.Raw)
                                                 )}
    elif "cryptography.hazmat.primitives.asymmetric.ed448" in sys.modules and isinstance(key,
                                                                                         ed448.Ed448PrivateKey):
        return "EdDSA", {"kty": "OKP", "crv": "Ed448",
                         "x": bytes_to_base64url(key.public_key().public_bytes(encoding=serialization.Encoding.Raw,
                                                                               format=serialization.PublicFormat.Raw)
                                                 )}
    else:
        raise ValueError("Unsupported key: {}".format(key))


# @brief sign string with key
def signature_of_str(key, string):
    alg, _ = get_key_alg_and_jwk(key)
    data = string.encode('utf8')
    if alg == 'RS256':
        return key.sign(data, padding.PKCS1v15(), hashes.SHA256())
    elif alg.startswith('ES'):
        full_octets = (int(alg[2:]) + 7) // 8
        if alg == 'ES256':
            der_sig = key.sign(data, ec.ECDSA(hashes.SHA256()))
        elif alg == 'ES384':
            der_sig = key.sign(data, ec.ECDSA(hashes.SHA384()))
        elif alg == 'ES512':
            der_sig = key.sign(data, ec.ECDSA(hashes.SHA512()))
        else:
            raise ValueError("Unsupported EC signature algorithm: {}".format(alg))
        # convert DER signature to RAW format (https://tools.ietf.org/html/rfc7518#section-3.4)
        r, s = decode_dss_signature(der_sig)
        return int_to_bytes(r, full_octets) + int_to_bytes(s, full_octets)
    elif alg == 'EdDSA':
        return key.sign(data)
    else:
        raise ValueError("Unsupported signature algorithm: {}".format(alg))


# @brief hash a string
def hash_of_str(string):
    account_hash = hashes.Hash(hashes.SHA256(), backend=default_backend())
    account_hash.update(string.encode('utf8'))
    return account_hash.finalize()


# @brief helper function to base64 encode for JSON objects
# @param b the byte-string to encode
# @return the encoded string
def bytes_to_base64url(b):
    return base64.urlsafe_b64encode(b).decode('utf8').replace("=", "")


# @brief check whether existing target file is still valid or source crt has been updated
# @param target string containing the path to the target file
# @param file string containing the path to the certificate file
# @return True if target file is at least as new as the certificate, False otherwise
def target_is_current(target, file):
    if not os.path.isfile(target):
        return False
    target_date = os.path.getmtime(target)
    crt_date = os.path.getmtime(file)
    return target_date >= crt_date


# @brief convert domain to idna representation (if applicable
def idna_convert(domain):
    try:
        if any(ord(c) >= 128 for c in domain):
            # Translate IDNA domain name from a unicode domain (handle wildcards separately)
            if domain.startswith('*.'):
                idna_domain = "*.{}".format(domain[2:].encode('idna').decode('ascii'))
            else:
                idna_domain = domain.encode('idna').decode('ascii')
            return idna_domain
    except Exception as e:
        log("Unicode domain(s) found but IDNA names could not be translated due to error: {}".format(e), error=True)
    return domain


# @brief validate the OCSP status for a given certificate by the given issuer
def is_ocsp_valid(cert, issuer, hash_algo):
    if hash_algo == 'sha1':
        algorithm = hashes.SHA1
    elif hash_algo == 'sha224':
        algorithm = hashes.SHA224
    elif hash_algo == 'sha256':
        algorithm = hashes.SHA256
    elif hash_algo == 'sha385':
        algorithm = hashes.SHA384
    elif hash_algo == 'sha512':
        algorithm = hashes.SHA512
    else:
        log("Invalid hash algorithm '{}' used for OCSP validation. Validation ignored.".format(hash_algo), warning=True)
        return True

    if isinstance(issuer, list):
        issuer = issuer[0]   # First certificate in the CA chain is the immediate issuer

    try:
        ocsp_urls = []
        aia = cert.extensions.get_extension_for_oid(ExtensionOID.AUTHORITY_INFORMATION_ACCESS)
        for data in aia.value:
            if data.access_method == x509.OID_OCSP:
                ocsp_urls.append(data.access_location.value)

        # This is a bit of a hack due to validation problems within cryptography (TODO: Check if this is still true)
        # Correct replacement:  ocsprequest = ocsp.OCSPRequestBuilder().add_certificate(cert, issuer, algorithm).build()
        ocsprequest = ocsp.OCSPRequestBuilder((cert, issuer, (algorithm)())).build()
        ocsprequestdata = ocsprequest.public_bytes(serialization.Encoding.DER)
        for ocsp_url in ocsp_urls:
            response = get_url(ocsp_url,
                               ocsprequestdata,
                               {
                                   'Accept': 'application/ocsp-response',
                                   'Content-Type': 'application/ocsp-request',
                               })
            ocspresponsedata = response.read()
            ocspresponse = ocsp.load_der_ocsp_response(ocspresponsedata)
            if ocspresponse.response_status == ocsp.OCSPResponseStatus.SUCCESSFUL \
                    and ocspresponse.certificate_status == ocsp.OCSPCertStatus.REVOKED:
                return False
    except Exception as e:
        log("An exception occurred during OCSP validation (Validation will be ignored): {}".format(e), error=True)

    return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# dns.nsupdate - rfc2136 based challenge handler
# Copyright (c) Rudolf Mayerhofer, 2019
# available under the ISC license, see LICENSE
import io
import re

import dns
import dns.query
import dns.tsigkeyring
import dns.update

from acertmgr.modes.dns.abstract import DNSChallengeHandler, QUERY_TIMEOUT
from acertmgr.tools import log

DEFAULT_KEY_ALGORITHM = "HMAC-MD5.SIG-ALG.REG.INT"


class ChallengeHandler(DNSChallengeHandler):
    @staticmethod
    def _read_tsigkey(tsig_key_file, key_name=None):
        try:
            with io.open(tsig_key_file) as key_file:
                key_struct = key_file.read()
                if not key_name:
                    key_name = re.search(r"key \"?([^\"{ ]+?)\"? {.*};", key_struct, re.DOTALL).group(1)
                key_data = re.search(r"key \"?%s\"? {(.*?)};" % key_name, key_struct, re.DOTALL).group(1)
                algorithm = re.search(r"algorithm ([a-zA-Z0-9_-]+?);", key_data, re.DOTALL).group(1)
                tsig_secret = re.search(r"secret \"(.*?)\"", key_data, re.DOTALL).group(1)
        except IOError as exc:
            raise ValueError("A problem was encountered opening your keyfile '{}': {}".format(tsig_key_file, exc))
        except AttributeError as exc:
            raise ValueError("Unable to decipher data from your keyfile: {}".format(exc))

        keyring = dns.tsigkeyring.from_text({
            key_name: tsig_secret
        })

        if not algorithm:
            algorithm = DEFAULT_KEY_ALGORITHM

        return keyring, algorithm

    def __init__(self, config):
        DNSChallengeHandler.__init__(self, config)
        if 'nsupdate_keyfile' in config:
            nsupdate_keyname = config.get("nsupdate_keyname", None)
            self.keyring, self.keyalgorithm = self._read_tsigkey(config.get("nsupdate_keyfile"), nsupdate_keyname)
        else:
            self.keyring = dns.tsigkeyring.from_text({
                config.get("nsupdate_keyname"): config.get("nsupdate_keyvalue")
            })
            self.keyalgorithm = config.get("nsupdate_keyalgorithm", DEFAULT_KEY_ALGORITHM)
        self.nsupdate_server = config.get("nsupdate_server")
        self.nsupdate_verify = config.get("nsupdate_verify", "true") == "true"
        self.nsupdate_verified = False

    def _determine_zone_and_nameserverip(self, domain):
        nameserver = self.nsupdate_server
        if nameserver:
            nameserverip = self._lookup_ip(nameserver)
            zone, _ = self._lookup_zone(domain, nameserverip)
        else:
            zone, nameserver = self._lookup_zone(domain)
            nameserverip = self._lookup_ip(nameserver)
        return zone, nameserverip

    def add_dns_record(self, domain, txtvalue):
        zone, nameserverip = self._determine_zone_and_nameserverip(domain)
        update = dns.update.Update(zone, keyring=self.keyring, keyalgorithm=self.keyalgorithm)
        update.add(domain, self.dns_ttl, dns.rdatatype.TXT, txtvalue)
        log('Adding \'{} {} IN TXT "{}"\' to {}'.format(domain, self.dns_ttl, txtvalue, nameserverip))
        dns.query.tcp(update, nameserverip, timeout=QUERY_TIMEOUT)

    def remove_dns_record(self, domain, txtvalue):
        zone, nameserverip = self._determine_zone_and_nameserverip(domain)
        update = dns.update.Update(zone, keyring=self.keyring, keyalgorithm=self.keyalgorithm)
        update.delete(domain, dns.rdata.from_text(dns.rdataclass.IN, dns.rdatatype.TXT, txtvalue))
        log('Deleting \'{} {} IN TXT "{}"\' from {}'.format(domain, self.dns_ttl, txtvalue, nameserverip))
        dns.query.tcp(update, nameserverip, timeout=QUERY_TIMEOUT)

    def verify_dns_record(self, domain, txtvalue):
        if self.nsupdate_verify and not self.dns_verify_all_ns and not self.nsupdate_verified:
            # Verify master DNS only if we don't do a full NS check and it has not yet been verified
            _, nameserverip = self._determine_zone_and_nameserverip(domain)
            if self._check_txt_record_value(domain, txtvalue, nameserverip, use_tcp=True):
                log('Verified \'{} {} IN TXT "{}"\' on {}'.format(domain, self.dns_ttl, txtvalue, nameserverip))
                self.nsupdate_verified = True
            else:
                # Master DNS verification failed. Return immediately and try again.
                return False

        return DNSChallengeHandler.verify_dns_record(self, domain, txtvalue)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# authority - authority api package
# Copyright (c) Rudolf Mayerhofer, 2019.
# available under the ISC license, see LICENSE

import importlib
import json
import os

from acertmgr import tools
from acertmgr.tools import log

authorities = dict()


# @brief find or create a suitable authority for the given settings
# @param settings the authority configuration options
def authority(settings):
    key = json.dumps(settings, sort_keys=True)
    if key in authorities:
        return authorities[key]
    else:
        acc_file = settings['account_key']
        if os.path.isfile(acc_file):
            log("Reading account key from {}".format(acc_file))
            acc_key = tools.read_pem_file(acc_file, key=True)
        else:
            log("Account key not found at '{0}'. Creating key.".format(acc_file))
            acc_key = tools.new_account_key(acc_file, settings['account_key_algorithm'], settings['account_key_length'])

        authority_module = importlib.import_module("acertmgr.authority.{0}".format(settings["api"]))
        authority_class = getattr(authority_module, "ACMEAuthority")
        authority_obj = authority_class(settings, acc_key)
        authorities[key] = authority_obj
        return authority_obj



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Automated Certificate Manager using ACME
# Copyright (c) Markus Hauschild & David Klaftenegger, 2016.
# Copyright (c) Rudolf Mayerhofer, 2019.
# available under the ISC license, see LICENSE

import io
import os
import stat
import subprocess
import sys

from acertmgr import configuration, tools
from acertmgr.authority import authority
from acertmgr.modes import challenge_handler
from acertmgr.tools import log, LOG_REPLACEMENTS

try:
    import pwd
    import grp
except ImportError:
    # Warnings will be reported upon usage below
    pass


# @brief fetch new certificate from letsencrypt
# @param settings the domain's configuration options
def cert_get(settings):
    log("Getting certificate for %s" % settings['domainlist'])

    acme = authority(settings['authority'])
    acme.register_account()

    # create challenge handlers for this certificate
    challenge_handlers = dict()
    for domain in settings['domainlist']:
        # Create the challenge handler
        challenge_handlers[domain] = challenge_handler(settings['handlers'][domain])

    # create ssl key
    key_file = settings['key_file']
    if os.path.isfile(key_file):
        key = tools.read_pem_file(key_file, key=True)
    else:
        log("SSL key not found at '{0}'. Creating key.".format(key_file))
        key = tools.new_ssl_key(key_file, settings['key_algorithm'], settings['key_length'])

    # create ssl csr
    csr_file = settings['csr_file']
    if os.path.isfile(csr_file) and str(settings['csr_static']).lower() == 'true':
        log('Loading CSR from {}'.format(csr_file))
        cr = tools.read_pem_file(csr_file, csr=True)
    else:
        log('Generating CSR for {}'.format(settings['domainlist']))
        must_staple = str(settings.get('cert_must_staple')).lower() == "true"
        cr = tools.new_cert_request(settings['domainlist'], key, must_staple)
        tools.write_pem_file(cr, csr_file)

    # request cert with csr
    crt, ca = acme.get_crt_from_csr(cr, settings['domainlist'], challenge_handlers, settings.get('cert_profile', None))

    #  if resulting certificate is valid: store in final location
    if tools.is_cert_valid(crt, settings['ttl_days']):
        log("Certificate '{}' renewed and valid until {}".format(tools.get_cert_identifier(crt),
                                                                 tools.get_cert_valid_until(crt)))
        tools.write_pem_file(crt, settings['cert_file'], stat.S_IREAD)
        if (not str(settings.get('ca_static')).lower() == 'true' or not os.path.exists(settings['ca_file'])) \
                and ca is not None:
            tools.write_pem_file(ca, settings['ca_file'])


# @brief put new certificate in place
# @param settings the domain's configuration options
# @return the action to be executed after the certificate update
def cert_put(settings):
    if 'path' not in settings:
        raise ValueError('Deployment settings are missing required element: path')
    if 'format' not in settings:
        raise ValueError('Deployment settings are missing required element: format')

    with io.open(settings['path'], "w+") as crt_fd:
        for fmt in [str.strip(x) for x in settings['format'].split(",")]:
            if fmt == "crt":
                with io.open(settings['cert_file'], "r") as src_fd:
                    crt_fd.write(src_fd.read())
            elif fmt == "key":
                with io.open(settings['key_file'], "r") as src_fd:
                    crt_fd.write(src_fd.read())
            elif fmt == "ca":
                with io.open(settings['ca_file'], "r") as src_fd:
                    crt_fd.write(src_fd.read())
            else:
                log("Ignored unknown deployment format key: {}".format(fmt), warning=True)

    # set owner and group
    if 'user' in settings or 'group' in settings:
        if 'pwd' in sys.modules and 'grp' in sys.modules and hasattr(os, 'chown') and hasattr(os, 'geteuid') and \
                hasattr(os, 'getegid'):
            try:
                uid = pwd.getpwnam(settings['user']).pw_uid if 'user' in settings else os.geteuid()
                gid = grp.getgrnam(settings['group']).gr_gid if 'group' in settings else os.getegid()
                os.chown(settings['path'], uid, gid)
            except OSError as e:
                log('Could not set certificate file ownership', e, warning=True)
        else:
            log('File user and group handling unavailable on this platform', warning=True)
    # set permissions
    if 'perm' in settings:
        if hasattr(os, 'chmod'):
            try:
                os.chmod(settings['path'], int(settings['perm'], 8))
            except OSError as e:
                log('Could not set certificate file permissions', e, warning=True)
        else:
            log('File permission handling unavailable on this platform', warning=True)

    return settings['action']


def cert_revoke(cert, configs, fallback_authority, reason=None):
    domains = set(tools.get_cert_domains(cert))
    acmeconfig = None
    for config in configs:
        if domains == set(config['domainlist']):
            acmeconfig = config['authority']
            break
    if not acmeconfig:
        acmeconfig = fallback_authority
        log("No matching authority found to revoke {}: {}, using globalconfig/defaults".format(tools.get_cert_identifier(cert),
                                                                                               tools.get_cert_domains(
                                                                                                   cert)), warning=True)
    acme = authority(acmeconfig)
    acme.register_account()
    acme.revoke_crt(cert, reason)


def cert_check_ari_for_renewal(cert, issuer, configs, fallback_authority):
    domains = set(tools.get_cert_domains(cert))
    acmeconfig = None
    for config in configs:
        if domains == set(config['domainlist']):
            acmeconfig = config['authority']
            break
    if not acmeconfig:
        acmeconfig = fallback_authority
        log("No matching authority found to revoke {}: {}, using globalconfig/defaults".format(tools.get_cert_identifier(cert),
                                                                                               tools.get_cert_domains(
                                                                                                   cert)), warning=True)
    acme = authority(acmeconfig)
    return acme.check_ari_for_renewal(cert, issuer)


def main():
    # load config
    runtimeconfig, domainconfigs = configuration.load()
    # register idna-mapped domains as LOG_REPLACEMENTS for better readability of log output
    for domainconfig in domainconfigs:
        LOG_REPLACEMENTS.update({k: "{} [{}]".format(k, v) for k, v in domainconfig['domainlist_idna_mapped'].items()})
    # Start processing
    if runtimeconfig.get('mode') == 'revoke':
        # Mode: revoke certificate
        log("Revoking {}".format(runtimeconfig['revoke']))
        cert_revoke(tools.read_pem_file(runtimeconfig['revoke']),
                    domainconfigs,
                    runtimeconfig['fallback_authority'],
                    runtimeconfig['revoke_reason'])
    else:
        # Mode: issue certificates (implicit)
        # post-update actions (run only once)
        actions = set()
        superseded = set()
        exceptions = list()
        # check certificate validity and obtain/renew certificates if needed
        for config in domainconfigs:
            try:
                # Load certificate if it already exists
                cert = None
                if os.path.isfile(config['cert_file']):
                    cert = tools.read_pem_file(config['cert_file'])
                # Load CA data necessary for issuer validation with OCSP and ARI
                issuer = None
                if cert and os.path.isfile(config['ca_file']):
                    try:
                        issuer = tools.read_pem_file(config['ca_file'])
                    except Exception as e1:
                        log("Failed to retrieve issuer from ca file: {}. Trying to download...".format(e1))
                        try:
                            issuer = tools.download_issuer_ca(cert)
                        except Exception as e2:
                            log("Failed to download issuer for cert file: {}. Cannot validate OCSP.".format(e2))
                if (not cert
                    or ('force_renew' in runtimeconfig
                        and all(d in config['domainlist'] for d in runtimeconfig['force_renew'])
                        )
                    or not tools.is_cert_valid(cert, config['ttl_days'])
                    or (str(config.get('validate_ocsp')).lower() != 'false'
                        and issuer and not tools.is_ocsp_valid(cert, issuer, config['validate_ocsp'])
                        )
                    or (str(config.get('validate_ari')).lower() != 'false' and issuer
                        and cert_check_ari_for_renewal(cert, issuer, domainconfigs, runtimeconfig['fallback_authority'])
                        )
                ):
                    cert_get(config)
                    if str(config.get('cert_revoke_superseded')).lower() == 'true' and cert:
                        superseded.add(cert)
            except Exception as e:
                log("Certificate issue/renew failed", e, error=True)
                exceptions.append(e)

        # deploy new certificates after all are renewed
        deployment_success = True
        for config in domainconfigs:
            for cfg in config['actions']:
                try:
                    if not tools.target_is_current(cfg['path'], config['cert_file']):
                        actions.add(cert_put(cfg))
                        log("Updated '{}' due to newer version".format(cfg['path']))
                except Exception as e:
                    log("Certificate deployment to {} failed".format(cfg['path']), e, error=True)
                    exceptions.append(e)
                    deployment_success = False

        # run post-update actions
        for action in actions:
            if action is not None:
                try:
                    # Run actions in a shell environment (to allow shell syntax) as stated in the configuration
                    output = subprocess.check_output(action, shell=True, stderr=subprocess.STDOUT)
                    logmsg = "Action succeeded: {}".format(action)
                    if len(output) > 0:
                        if getattr(output, 'decode', None):
                            # Decode function available? Use it to get a proper str
                            output = output.decode('utf-8')
                        logmsg += os.linesep + tools.indent(output, 18)  # 18 = len("Action succeeded: ")
                    log(logmsg)
                except subprocess.CalledProcessError as e:
                    output = e.output
                    logmsg = "Action failed: ({}) {}".format(e.returncode, e.cmd)
                    if len(output) > 0:
                        if getattr(output, 'decode', None):
                            # Decode function available? Use it to get a proper str
                            output = output.decode('utf-8')
                        logmsg += os.linesep + tools.indent(output, 15)  # 15 = len("Action failed: ")
                    log(logmsg, error=True)
                    exceptions.append(e)
                    deployment_success = False

        # revoke old certificates as superseded
        if deployment_success:
            for superseded_cert in superseded:
                try:
                    log("Revoking '{}' valid until {} as superseded".format(
                        tools.get_cert_identifier(superseded_cert),
                        tools.get_cert_valid_until(superseded_cert)))
                    cert_revoke(superseded_cert, domainconfigs, runtimeconfig['fallback_authority'], reason=4)
                except Exception as e:
                    log("Certificate supersede revoke failed", e, error=True)
                    exceptions.append(e)

        # throw a RuntimeError with all exceptions caught while working if there were any
        if len(exceptions) > 0:
            raise RuntimeError("{} exception(s) occurred during processing".format(len(exceptions)))



import io
import os
import subprocess

from setuptools import setup, find_packages


# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(filename):
    with io.open(os.path.join(os.path.dirname(__file__), filename)) as f:
        return f.read()


# Utility function to determine version using git in a PEP-440 compatible way, fallback to version.txt for releases
def determine_version():
    dir_path = os.path.dirname(os.path.realpath(__file__))
    ver_file = os.path.join(dir_path, "version.txt")
    version = "0.0.0"
    if os.path.exists(ver_file):
        version = read(ver_file)
    # If this is a release file and no git is found, use version.txt
    if not os.path.isdir(os.path.join(dir_path, ".git")):
        return version
    # Derive version from git
    try:
        output = subprocess.check_output(["git", "describe", "--tags", "--dirty"], cwd=dir_path) \
            .decode("utf-8").strip().split("-")
        if len(output) == 1:
            return output[0]
        elif len(output) == 2:
            return "{}.dev0".format(output[0])
        else:
            release = "dev" if len(output) == 4 and output[3] == "dirty" else ""
            return "{}.{}{}+{}".format(output[0], release, output[1], output[2])
    except subprocess.CalledProcessError:
        try:
            commit = subprocess.check_output(["git", "rev-parse", "HEAD"]).decode("utf-8").strip()
            status = subprocess.check_output(["git", "status", "-s"]).decode("utf-8").strip()
            return "{}.dev0+{}".format(version, commit) if len(status) > 0 else "{}+{}".format(version, commit)
        except subprocess.CalledProcessError:
            # finding the git version has utterly failed, use version.txt
            return version


extra_requirements = {
    "dns": ["dnspython"],
    "yaml": ["PyYAML"],
    "idna": ["idna"],
    "ocsp-must-staple": ["cryptography>=2.1"],
    "ocsp-validation": ["cryptography>=2.4"],
    "ed25519": ["cryptography>=2.6"],
}

if __name__ == "__main__":
    setup(
        name="acertmgr",
        version='1.1.1',
        author="Markus Hauschild",
        author_email="moepman@binary-kitchen.de",
        description="An automated certificate manager using ACME/letsencrypt",
        license="ISC",
        keywords="acme letsencrypt",
        url="https://github.com/moepman/acertmgr",
        packages=find_packages(),
        long_description=read("README.md"),
        long_description_content_type="text/markdown",
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Programming Language :: Python",
            "Environment :: Console",
            "Topic :: Security :: Cryptography",
            "License :: OSI Approved :: ISC License (ISCL)",
        ],
        install_requires=[
            "cryptography>=0.6",
        ],
        extras_require=extra_requirements,
        entry_points={
            "console_scripts": [
                "acertmgr=acertmgr:main",
            ],
        },
        data_files=[("readme", ["README.md"])],
    )
